"Fixture for setting up the tests, yield CliRunner() and tearing them down"
import os
import shutil
from collections import OrderedDict
from pathlib import Path

import fiona
import numpy as np
import pytest
import rasterio
from classifier.settings import WORKSPACE
from click.testing import CliRunner


@pytest.fixture
def runner(test_arg):
    """Setup, yield CliRunner() and teardown"""
    # setup
    print("SETUP")
    is_timeseries_test = test_arg[0]
    is_supervised_test = test_arg[1]
    # Rasters
    nowhere = 0.0
    pixel_size = 2.0

    transform = rasterio.Affine(
        pixel_size,
        nowhere,
        nowhere,
        nowhere,
        -pixel_size,
        nowhere
    )

    raster_folder = Path(WORKSPACE) / "integration_test_rasters"
    raster_folder.mkdir()

    if is_timeseries_test:
        raster_names = ['test_raster_008570_2020-07-01T000000Z_mean.tif',
                        'test_raster_008570_2020-08-01T000000Z_mean.tif']
    else:
        raster_names = ['test_raster_008570_2020-07-01T000000Z_mean.tif']
    raster_paths = [raster_folder / raster_name
                    for raster_name in raster_names]

    raster_values = [
        np.array([
            np.full((10, 10), [1, 1, 1, 1, 1, 100, 100, 100, 100, 100]),
            np.full((10, 10), [2, 2, 2, 2, 2, 101, 101, 101, 101, 101])
        ], dtype=np.uint8),
        np.array([
            np.full((10, 10), [1, 1, 1, 1, 1, 100, 100, 100, 100, 100]),
            np.full((10, 10), [2, 2, 2, 2, 2, 101, 101, 101, 101, 101])
        ], dtype=np.uint8),
    ]

    for raster_path, data in zip(raster_paths, raster_values):
        with rasterio.open(
            raster_path,
            'w',
            driver='GTiff',
            height=data.shape[1],
            width=data.shape[2],
            count=data.shape[0],
            dtype='uint8',
            crs='+proj=latlong',
            transform=transform
        ) as dst:
            dst.write(data)

    if is_supervised_test:
        # Train polygons
        x_min, y_min, x_max, y_max = 0, -20, 10, 0
        window_polygon_1 = [[(x_min, y_min),
                             (x_min, y_max),
                             (x_max, y_max),
                             (x_max, y_min)]]

        x_min, y_min, x_max, y_max = 10, -20, 20, 0
        window_polygon_2 = [[(x_min, y_min),
                             (x_min, y_max),
                             (x_max, y_max),
                             (x_max, y_min)]]

        polygon1 = {
            "properties": {
                "id": 0,
                "fid": 1
            },
            "geometry": {
                "coordinates": window_polygon_1,
                "type": "Polygon"
            }
        }
        polygon2 = {
            "type": "Feature",
                    "properties": {
                        "id": 1,
                        "fid": 2
                    },
            "geometry": {
                        "coordinates": window_polygon_2,
                        "type": "Polygon"
                    }
        }

        train_polygons = [polygon1, polygon2]
        schema = {
            "geometry": "Polygon",
            "properties": OrderedDict([("id", "int"), ("fid", "int")])
        }
        rois_path = Path(WORKSPACE) / 'integration_test_rois.gpkg'
        with fiona.open(rois_path,
                        "w",
                        driver="GPKG",
                        crs="+proj=latlong",
                        schema=schema) as src:
            src.writerecords(train_polygons)

    yield CliRunner()

    # teardown
    shutil.rmtree(Path(WORKSPACE) / "integration_test_rasters")
    if is_supervised_test:
        os.remove(Path(WORKSPACE) / 'integration_test_rois.gpkg')
    # Delete classification result
    shutil.rmtree(Path(WORKSPACE) / "test_result")



"""Settings file for Classifier"""
from sklearn.cluster import KMeans, MiniBatchKMeans

from classifier.models import RandomForest, SingleClass, XGBoost

WORKSPACE = '/workspace/'

RASTER_EXTENSIONS = frozenset(['.tif', '.vrt', '.jp2'])

US_ALGORITHMS = ['us_kmeans', 'us_kmeans_minibatch']

# ##----THE ALGORITHMS---###
ALGORITHMS = [
    "randomforest", "xgboost", "singleclass", "us_kmeans",
    "us_kmeans_minibatch"
]
CLASSIFIERS = [RandomForest, XGBoost, SingleClass, KMeans, MiniBatchKMeans]
ALGORITHM_DICT = dict(zip(ALGORITHMS, CLASSIFIERS))

PARAMETERS = {
    "app": {
        "algorithm": "randomforest",
        "window": 1024,
        "model": None,
        "model_save": False,  # Save a model file which can be re-used
        "samples": None,
        "log_level": "INFO",
        "threads": -1,
        "imputation": True,
        "imputation_strategy": "mean",
        "imputation_constant": -9999,
        "rasters_are_timeseries": False
    },
    "supervised": {
        "probability": True,
        "all_probabilities": False,
        "boxplots": False,  # Plot Boxplots for samples
        "remove_outliers": True,  # Remove outliers from the training data
        "optimization": {
            "optimize": False,  # Optimize the model parameters
            "optimize_number": 10,  # Number of iterations for optimization
            "max_features": ['auto', 'sqrt', 'log2'],
            "max_leaf_nodes": [3, 5, 7],
            "max_depth": [None, 1, 3, 10, 20000],
            "n_estimators": [10, 50, 100]
        }
    },
    "unsupervised": {
        "nclasses": 2,  # Number of classes for unsupervised
        "trainfraction": 1.0  # Fraction of raster used for training
    },
    "accuracy": {
        "perform_assesment": True,  # Perform accuracy assessment
        "testfraction": 0.25,  # Fraction of data to use for training
    }
}

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
IMPUTATION_STRATEGIES = ["mean", "median",
                         "most_frequent", "constant", "interpolate"]

OPTIMIZE_RANDOMFOREST_MAX_FEATURES = ["auto", "sqrt", "log2"]



"""An sk-learn classifier for landcover classification"""

import logging
import os
import shutil
import sys
import time


import fiona
import pandas as pd

from classifier.accuracy import (plot_feature_importances,
                                 write_confusion_matrix)
from classifier.dataprep import createdataset, outlier_removal
from classifier.predict import prediction
from classifier.settings import US_ALGORITHMS, WORKSPACE
from classifier.timeseries import get_timeseries_samples
from classifier.train import train_dataset
from classifier.unsupervised import make_input_array, train_kmeans
from classifier.utils.general import parallel_function, read_model, save_model
from classifier.utils.raster import get_meta, stitch
from classifier.utils.config import Configuration

START = time.time()
CLASSIFIER_LOGGER = logging.getLogger(__name__)


def read_samples(samples, remove_outliers=True):
    """Read a samples file.

    Args:
        samples(path): CSV file containing samples
        remove_outliers (Bool): Flag for removing outliers from samples

    Returns:
        sample_df(DataFrame): DF containing samples

    """
    sample_df = pd.read_csv(samples, index_col=0)
    if remove_outliers:
        sample_df = outlier_removal(sample_df)
    return sample_df


def get_rois_extent(rois):
    """

    Args:
        rois (str): path to a rois file

    Returns:
        bounds of rois (list): ulx, uly, llx, lly coordinates

    """
    with fiona.open(rois, "r") as shapefile:
        return shapefile.bounds


def gather_samples(rois, rasters, out_dir, config):
    """Gather samples from a rois file and raster files

    Args:
        rois: The path to a rois file
        rasters: A list of raster files
        out_dir: Path where to write the dataset file
        config (Configuration) contains config

    Returns:
        A dataframe containing pixel values and classes

    """
    # Train a model and continue
    samples = createdataset(
        rasters,
        rois,
        out_dir,
        config)
    samples.to_csv(os.path.join(out_dir, 'samples.csv'))
    return samples


def train(samples, rasters, out_dir, config, rois_extent):
    """Train a model using samples and rasters

    Args:
        samples: Samples dataframe containing pixel values and class values
        rasters: A list of raster files
        out_dir: Path where to write the dataset file
        config (Configuration) contains config
        rois_extent: The extent of the training data (if any, else None)

    Returns:
        The return value. True for success, False otherwise.

    """
    CLASSIFIER_LOGGER.info("\n####-----Training----#####\n")
    dataset = samples
    windows, _ = get_meta(rasters, config.app.window)
    if config.app.algorithm in US_ALGORITHMS and len(samples) < 1:
        # Do unsupervised
        array = make_input_array(
            rasters,
            windows,
            config
        )
        model_dict = train_kmeans(array,
                                  config)
    else:  # All supervised methods
        model_dict, test = train_dataset(
            dataset,
            config.app.algorithm,
            out_dir,
            config
        )
        if config.app.algorithm in ['randomforest', 'xgboost'] \
                and config.accuracy.perform_assesment:
            # Do the accuracy analysis
            cm_fn = os.path.join(out_dir, 'confusion_matrix')
            fi_fn = os.path.join(out_dir, 'feature_importance.png')
            write_confusion_matrix(model_dict, test, cm_fn)
            plot_feature_importances(model_dict, fi_fn)
    # save the model as a python pickle
    if not rois_extent is None:
        model_dict['rois_bounds'] = rois_extent
    if config.app.model_save:
        save_model(model_dict, out_dir, config)
    CLASSIFIER_LOGGER.info("\nFinished Training\n")
    return model_dict


def predict(model_dict, rasters, out_dir, config):
    """Prediction function using a trained model and raster files

    Args:
        model_dict: A trained model
        rasters: A list of raster files
        out_dir: Path where to write the dataset file
        config (Configuration) contains config

    """
    threads = config.app.threads
    CLASSIFIER_LOGGER.info("\n####-----Prediction----#####\n")
    windows, meta = get_meta(rasters, config.app.window)

    iterable = [
        {
            'window': window[1],
            'rasters': rasters,
            'model_dict': model_dict,
            'meta': meta,
            'config': config
        }
        for window in windows
    ]
    if threads > 1 or threads == -1:
        parallel_function(prediction,
                          iterable,
                          threads)
    else:
        for wins in iterable:
            prediction(**wins)

    # ##--------------------STITCHING-----------------------------------###

    CLASSIFIER_LOGGER.info("\n####-----Stitching----#####\n")

    # Run the gdalwarp command in the command line
    stitch(out_dir, meta, config)
    CLASSIFIER_LOGGER.info("Cleaning up..")
    shutil.rmtree(config.tmp_dir)
    CLASSIFIER_LOGGER.info(
        "Total run time was %i seconds", (int(time.time() - START)))


def train_and_predict_with_samples(samples, rasters, out_dir, config,
                                   rois_extent=None):
    """Train model using samples

    Args:
        samples: Samples dataframe containing pixel values and class values
        rasters: A list of raster files
        out_dir: Path where to write the dataset file
        config (Configuration) contains config

    Returns:
        The return value. True for success, False otherwise.

    """
    model_dict = train(samples, rasters, out_dir, config, rois_extent)
    predict(model_dict, rasters, out_dir, config)


def gather_train_predict_rois(rois, rasters, out_dir, config):
    """Gather samples from rois file and rasters and continue with training and
    prediction

    Args:
        rois: The path to a rois file
        rasters: A list of raster files
        out_dir: Path where to write the dataset file
        config (Configuration) contains config

    Returns:
        The return value. True for success, False otherwise.

    """
    samples = gather_samples(rois, rasters, out_dir, config)
    rois_extent = get_rois_extent(rois)
    return train_and_predict_with_samples(samples,
                                          rasters,
                                          out_dir,
                                          config,
                                          rois_extent)


def classify_(rasters, config, out_dir, rois=None):
    """Entry point function for pixel-based classification

        Args:
            rasters (list): List of raster paths
            config (Configuration): CLI configuration dict
            out_dir (str): Path to output directory
            rois (str): Path to rois file

    """
    if config.app.model is not None:
        model = read_model(config.app.model)
        predict(model, rasters, out_dir, config)
    if config.app.algorithm not in US_ALGORITHMS:
        print(config.app.algorithm)
        # Supervised classification
        if config.app.samples is not None:

            samples = read_samples(
                config.app.samples,
                config.supervised.remove_outliers)

            train_and_predict_with_samples(samples,
                                           rasters,
                                           out_dir,
                                           config)

        elif rois is not None:
            gather_train_predict_rois(
                rois,
                rasters,
                out_dir,
                config)
        else:
            CLASSIFIER_LOGGER.error(
                "You have chosen a supervised classification method but didn't "
                "provide rois or samples. Either provide one of these or run a "
                "unsupervised classification")
            sys.exit(1)
    else:
        print("eHEEERE")
        if config.app.samples is not None or rois is not None:
            CLASSIFIER_LOGGER.error(
                "You have chosen an unsupervised classification "
                "method but provided samples or rois. Please choose "
                "a supervised method to use them or don't provide them for "
                "an unsupervised classification.")
            sys.exit(1)
        CLASSIFIER_LOGGER.info("Doing unsupervised classification")
        train_and_predict_with_samples([],
                                       rasters,
                                       out_dir,
                                       config)


def classify_timeseries(rasters: list,
                        config: Configuration,
                        out_dir: str,
                        rois: str):
    """Main function to classifiy timeseries.

    For now, only from start to finish is supported. ie, supplying a model
    or samples does not work.

        Args:
            rasters (Configuration): contains config
            out_dir (str): Path to output directory
            rois (str): Path to rois file

    """
    # Gather samples and do imputation.
    CLASSIFIER_LOGGER.info("Now Getting Samples and doing imputation")
    if rois is None:
        CLASSIFIER_LOGGER.error(
            "You have to provide rois for timeseries classification.")
        sys.exit()
    rois = os.path.join(WORKSPACE, rois)
    samples_df = get_timeseries_samples(rasters, rois, out_dir, config)
    # Reshape it
    samples = samples_df.T
    samples['class'] = samples.index
    if config.app.algorithm in US_ALGORITHMS:
        CLASSIFIER_LOGGER.error(
            "Unsupervised classification not supported yet")
        sys.exit()
    model = train(samples,
                  rasters=rasters,
                  out_dir=out_dir,
                  config=config,
                  rois_extent=None)

    predict(model, rasters, out_dir, config)


if __name__ == "__main__":
    pass



"Supervised single raster integration test for classifier"
from pathlib import Path

import numpy as np
import pytest
import rasterio
from classifier.cli import classification
from classifier.settings import WORKSPACE
from classifier.utils.general import save_config_as_json


@pytest.mark.parametrize('test_arg',
                         [[False, True]])
def test_supervised_singleraster_randomforest(runner):
    """
    Test for running a supervised randomforest classification
    on a single raster
    """
    save_config_as_json(threads=1, algorithm="randomforest",
                        remove_outliers=False)

    rois_path = Path(WORKSPACE) / 'integration_test_rois.gpkg'
    runner.invoke(classification, [
        "--name", 'test_result', '--overwrite', True,
        '--rois', rois_path, "integration_test_rasters"])
    directory = Path(WORKSPACE) / 'test_result' / 'classification.tif'
    with rasterio.open(directory, 'r') as dst:
        result = dst.read(1)
        assert np.all(result[:, 0:4] == result[0, 0])
        assert np.all(result[:, 5:10] == result[0, 5])


@pytest.mark.parametrize('test_arg',
                         [[False, True]])
def test_supervised_singleraster_xgboost(runner):
    """
    Test for running a supervised xgboost classification
    on a single raster
    """
    save_config_as_json(threads=1, algorithm="xgboost",
                        remove_outliers=False)

    rois_path = Path(WORKSPACE) / 'integration_test_rois.gpkg'
    runner.invoke(classification, [
        "--name", 'test_result', '--overwrite', True,
        '--rois', rois_path, "integration_test_rasters"])
    directory = Path(WORKSPACE) / 'test_result' / 'classification.tif'
    with rasterio.open(directory, 'r') as dst:
        result = dst.read(1)
        assert np.all(result[:, 0:4] == result[0, 0])
        assert np.all(result[:, 5:10] == result[0, 5])


@pytest.mark.parametrize('test_arg',
                         [[False, True]])
def test_supervised_singleraster_singleclass(runner):
    """
    Test for running a supervised singleclass classification
    on a single raster
    """
    save_config_as_json(threads=1, algorithm="singleclass",
                        remove_outliers=False)

    rois_path = Path(WORKSPACE) / 'integration_test_rois.gpkg'
    runner.invoke(classification, [
        "--name", 'test_result', '--overwrite', True,
        '--rois', rois_path, "integration_test_rasters"])
    directory = Path(WORKSPACE) / 'test_result' / 'classification.tif'
    with rasterio.open(directory, 'r') as dst:
        result = dst.read(1)
        assert np.all(result[:, 0:4] == result[0, 0])
        assert np.all(result[:, 5:10] == result[0, 5])



"Accuracy assessment and plotting for classifier"
import itertools
import logging
import time

import fiona
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio
from rasterio import mask
from shapely.geometry import shape
from shapely.ops import cascaded_union
from sklearn.metrics import cohen_kappa_score, confusion_matrix

ACCURACY_LOGGER = logging.getLogger(__name__)


def write_confusion_matrix(model_dict, test_set, cm_fn, plot=True, csv=True):
    """ Uses a sklearn confusion matrix (np 2d array) to write to a csv file
    that also contains overall metrics. Also plots the confusion matrix for
    easier viewing.

        Args:
            model_dict (dict):dictionary with the name, model and label encoder
            test_set (pd.DataFrame): test dataset not used during training
            cm_fn (str): The filename of the output csv
            plot (str): Whether or not to plot the figure
            csv (str): Whether or not to write csv file

        returns:
            nothing
    """
    # Convert dict and dataframe to arrays
    x_test = test_set[[x for x in test_set.columns if 'class' not in x]].values
    y_test = np.ravel(
        test_set[[x for x in test_set.columns if 'class' in x]].values)
    preds = model_dict['model'].predict(x_test)
    compute_confusion_matrix(y_test, preds, cm_fn, plot, csv)


def compute_confusion_matrix(y_test, preds, cm_fn, plot=True, csv=True):
    """ Uses a sklearn confusion matrix (np 2d array) to write to a csv file
    that also contains overall metrics. Also plots the confusion matrix for
    easier viewing.

        Args:
            y_test (np.array): Ground truth (correct) target values
            preds (np.array): Estimated targets as returned by a classifier
            cm_fn (str): The filename of the output
            plot (str): Whether or not to plot the figure
            csv (str):  Whether or not to write the csv
    """

    # Write and plot confusion Matrix
    cm_labels = sorted(list(set(np.unique(y_test)) | set(np.unique(
        preds))))
    kappa = cohen_kappa_score(y_test, preds, labels=cm_labels)
    conf_matrix = confusion_matrix(y_test, preds, labels=cm_labels)

    # Metrics: overall mean accuracy and kappa
    ACCURACY_LOGGER.info("\n####-----Accuracy Assessment-----#####\n")
    metrics_list = []
    metrics_list.append(
        f'Overall Accuracy {np.trace(conf_matrix)/np.nansum(conf_matrix)}'
    )
    metrics_list.append(f'Kappa {kappa:.4f}')
    for metric in metrics_list:
        ACCURACY_LOGGER.info(metric)

    if csv:
        df_cm = pd.DataFrame(
            data=conf_matrix, index=cm_labels, columns=cm_labels)
        # Total of the rows
        df_cm['Total'] = df_cm.sum(axis=1)
        # Accuracy
        df_cm['Accuracy'] = np.diag(df_cm[cm_labels]) / df_cm['Total']
        # Appending the numbers to the df
        df_cm = df_cm.append(df_cm.sum(axis=0).rename('Total')).append(
            (pd.Series(np.diag(conf_matrix) / np.sum(conf_matrix, axis=0),
                       index=cm_labels).rename('Reliability')))
        df_cm.to_csv(cm_fn+'.csv', sep=';', float_format='%.4f')

    if plot:
        # Plotting if necessary
        try:
            # Plot normalized confusion matrix
            plt.figure()
            plot_confusion_matrix(conf_matrix,
                                  classes=cm_labels,
                                  normalize=True
                                  )
            plt.savefig(cm_fn + '_plot.png', dpi=150)
        except AttributeError:  # Not all models have FIs, so skip
            pass


def plot_confusion_matrix(conf_matrix, classes,
                          normalize=False,
                          cmap=plt.cm.Blues,
                          labeldict=None):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """
    if normalize:
        conf_matrix = conf_matrix.astype('float') /\
            conf_matrix.sum(axis=1)[:, np.newaxis]
    if labeldict is not None:
        classes = map(labeldict.get, classes)
    plt.imshow(conf_matrix, interpolation='nearest', cmap=cmap)
    plt.title("Confusion Matrix")
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=90, fontsize=6)
    plt.yticks(tick_marks, classes, fontsize=8)

    fmt = '.2f' if normalize else 'd'
    thresh = conf_matrix.max() / 2.
    for i, j in itertools.product(range(conf_matrix.shape[0]),
                                  range(conf_matrix.shape[1])):
        plt.text(j, i, format(conf_matrix[i, j], fmt),
                 horizontalalignment="center",
                 color="white" if conf_matrix[i, j] > thresh else "black",
                 fontsize=4)

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')


def plot_feature_importances(model_dict, outfile):
    """ Plot the feature importances of the forest"""
    importances = model_dict['model'].feature_importances_
    indices = np.argsort(importances)[::-1]
    labels = [model_dict['names'][x] for x in indices]
    nr_features = len(model_dict['names'])
    fig, axis = plt.subplots()
    axis.set_title("Feature importances (band numbers)")
    # Get STD whenever possible
    if model_dict['app_algorithm'] == 'randomforest':
        std = np.std([tree.feature_importances_ for tree in model_dict[
            'model'].estimators_], axis=0)
        axis.bar(range(nr_features), importances[indices],
                 color="r", yerr=std[indices], align="center")
    else:
        axis.bar(range(nr_features), importances[indices],
                 color="r", align="center")
    axis.set_xticks(range(nr_features))
    axis.set_xticklabels(labels, rotation=45, ha="right")
    axis.set_xlim([-1, nr_features])
    fig.tight_layout()
    plt.savefig(outfile, dpi=300)


def __gather_samples_for_roi(roi, rasterfile):
    """Get sample values from rasters

        Args:
            roi (fiona shape): region of interest
            raster (str): raster file name
            warp_dst (dict): warp specification

        Returns:
            np.array [nsamples, nfeatures] of samples from within rois.

    """
    with rasterio.open(rasterfile) as src:
        selection, _, window = mask.raster_geometry_mask(
            src,
            [roi["geometry"]],
            crop=True)
        raster_subset = src.read(window=window)

        # rasterio convention: outside shape=True, inside=False. We invert.
        samples_per_raster = raster_subset[:, ~selection]
        roi_values = np.asarray(samples_per_raster).flatten()

    return roi_values


def assess_accuracy(raster, rois, cm_fn, subset=None):
    """Accuracy assessment
        Args:
            raster (str): filename of classication raster
            rois (str): rois filename
            cm_fn (str): filename of output
            subset (str): filename of subset used for classication
    """

    start = time.time()

    # Is subset is given, exclude its polygons from accuracy assessment
    if subset:
        polygons = [
            shape(feature['geometry']) for feature in fiona.open(subset)]
        union_subset = cascaded_union(polygons)

    sample_labels = []
    all_samples = []
    counter = 0
    with fiona.open(rois, "r") as shapefile:
        for roi in shapefile:
            if subset and shape(roi['geometry']).intersects(union_subset):
                counter += 1
                continue
            roi_samples = __gather_samples_for_roi(
                roi,
                raster
            )
            if roi_samples is not None:
                all_samples += list(roi_samples.astype(np.float))
                roi_id = int(roi['properties']['id'])
                sample_labels += len(roi_samples) * [roi_id]
        all_samples = np.asarray(all_samples)
        sample_labels = np.asarray(sample_labels)

        ACCURACY_LOGGER.info(
            "Using %i groundtruth polygons, \
                excluding %i already used in classifier",
            len(shapefile),
            counter)

    # write confusion matrix
    compute_confusion_matrix(
        sample_labels, all_samples, cm_fn, plot=True, csv=False)

    ACCURACY_LOGGER.info(
        "Total run time was %i seconds", (int(time.time() - start)))



"Unsupervised classification test for classifier"
from pathlib import Path

import numpy as np
import pytest
import rasterio
from classifier.cli import classification
from classifier.settings import WORKSPACE
from classifier.utils.general import save_config_as_json


@pytest.mark.parametrize('test_arg',
                         [[False, False]])
def test_unsupervised_kmeans(runner):
    """
    Test for running an unsupervised kmeans classification
    """
    save_config_as_json(threads=1, algorithm="us_kmeans")

    runner.invoke(classification, [
        "--name", 'test_result', '--overwrite', True,
        "integration_test_rasters"])
    directory = Path(WORKSPACE) / 'test_result' / 'classification.tif'
    with rasterio.open(directory, 'r') as dst:
        result = dst.read(1)
        assert np.all(result[:, 0:4] == result[0, 0])
        assert np.all(result[:, 5:10] == result[0, 5])


@pytest.mark.parametrize('test_arg',
                         [[False, False]])
def test_unsupervised_kmeans_minibatch(runner):
    """
    Test for running an unsupervised kmeans-minibatch classification
    """
    save_config_as_json(threads=1, algorithm="us_kmeans_minibatch")

    runner.invoke(classification, [
        "--name", 'test_result', '--overwrite', True,
        "integration_test_rasters"])
    directory = Path(WORKSPACE) / 'test_result' / 'classification.tif'
    with rasterio.open(directory, 'r') as dst:
        result = dst.read(1)
        assert np.all(result[:, 0:4] == result[0, 0])
        assert np.all(result[:, 5:10] == result[0, 5])



# -*- coding: utf-8 -*-


# (c) Satelligence, see LICENSE.


# pylint: skip-file


from setuptools import setup


import setuptools


import os





version = '3.1.3'



long_description = open('README.md').read()





test_requirements = [


    'pytest'


]





setup(


    name='s11-classifier',


    version=version,


    description="Classifier",


    long_description=long_description,


    long_description_content_type="text/markdown",


    author="Satelligence",


    author_email='team@satelligence.com',


    url='https://gitlab.com/satelligence/classifier',


    packages=setuptools.find_packages(),


    package_dir={


        'classifier': 'classifier'


    },


    include_package_data=True,


    install_requires=[


        'boto3>=1.16.63, <2.0.0',


        'click>=7.1.2, <9.0.0',


        'dacite>=1.6.0, <2.0.0',


        'folium>=0.12.1, <1.0.0',


        'geopandas>=0.10.0, <1.0.0',


        'geojson>=2.5.0, <3.0.0',


        'marshmallow>=3.14.1, < 4.0.0',


        'matplotlib>=3.5.1, <4.0.0',


        'numpy>=1.22.2, < 2.0.0',


        'pylint>=2.12.0, <3.0.0',


        'pytest>=5.2.0, <6.0.0',


        'python-dateutil>=2.8.1, <3.0.0',


        'rasterio>=1.2.10, <2.0.0',


        'rasterstats>=0.15.0, <1.0.0',


        'rtree>=0.9.7, <1.0.0',


        'scikit_learn==1.0.2',


        'xarray>=0.12.3, <1.0.0',


        'xgboost>=1.1.1, <2.0.0',


    ],


    license="Apache-2.0",


    zip_safe=False,


    python_requires='>=3.5'


)





"""All timeseries functionality can be found here"""
import datetime
import logging
import sys
from multiprocessing import Manager

import fiona
import numpy as np
import pandas as pd

from classifier.dataprep import gather_samples_for_roi
from classifier.utils.general import impute_values, parallel_function
from classifier.utils.raster import (get_raster_date, raster_warp_dst,
                                     verify_and_count_bands)

TIMESERIES_LOGGER = logging.getLogger(__name__)


def gather_samples_ts_per_roi(
        mp_list, roi, ts_dict, config, warp_dst,
        df_index, bands):
    """Gather samples for a timeseries for a roi, for usage in MP function

        Args:
            mp_list (mp.List): list for use in multiprocessing
            roi (shapely.geometry): region of interest
            ts_dict(dict): dictionary of dates and rasters
            config (Configuration) contains config
            warp_dst (dict): warp parameters for raster
            df_index (list): index to use for creation of samples df
            bands (list): bands to use from the tifs
    """

    roi_samples = gather_samples_for_roi(
        roi,
        list(ts_dict.values()),
        warp_dst
    )

    index = pd.MultiIndex.from_product(
        [df_index, bands],
        names=['Date', 'Band']
    )
    roi_id = int(roi['properties']['id'])
    roi_classes = roi_samples.shape[0] * [roi_id]
    last_df = pd.DataFrame(
        roi_samples.transpose(),
        index=index,
        columns=roi_classes)

    imputed_df = impute_timeseries(last_df, config)
    mp_list.append(imputed_df)


def gather_samples_ts(ts_dict, config, rois_file=None, roi=None):
    """Gathers the pixel values for all the timeseries rasters and
    combines them in a df

    The DataFrame is a multiColumn dataframe where column level 0 is the class

        Args:
            ts_dict (dict): timeseries raster dictionary
            config (Configuration) contains config
            rois_file (str): Path to the ogr rois file when doing sample
                             gathering for training a model
            roi (shapely.geometry: geometry when gathering pixel values when
                                    predicting a single chunk

        returns:
            samples_df (DataFrame): DataFrame with timeseries of samples
    """

    # Get pandas index
    df_index = [datetime.datetime.strptime(x, "%Y-%m-%d") for x in ts_dict]
    # Iterate through the polygons and return
    rasters = list(ts_dict.values())
    warp_dst = raster_warp_dst(rasters[0])
    b_count = verify_and_count_bands(rasters)
    bands = [f'B{str(x).zfill(2)}' for x in np.arange(1, b_count + 1)]
    with Manager() as manager:
        mp_list = manager.list()
        if rois_file is not None:
            with fiona.open(rois_file, "r") as shapefile:
                args = [
                    {
                        'mp_list': mp_list,
                        'roi': roi,
                        'ts_dict': ts_dict,
                        'config': config,
                        'warp_dst': warp_dst,
                        'df_index': df_index,
                        'bands': bands
                    }
                    for roi in shapefile
                ]
        else:
            args = [
                {
                    'mp_list': mp_list,
                    'roi': roi,
                    'ts_dict': ts_dict,
                    'config': config,
                    'warp_dst': warp_dst,
                    'df_index': df_index,
                    'bands': bands
                }
            ]
        parallel_function(
            gather_samples_ts_per_roi,
            args,
            ncpus=config.app.threads
        )
        samples_df = pd.concat(mp_list, axis=1)
        samples_df.columns = samples_df.columns.astype(str)
        return samples_df


def impute_timeseries(samples_df, config):
    """Imputation of timeseries

        Args:
            samples_df(pd.DataFrame): Dataframe with timeseries samples
            config (Configuration) contains config

        Returns:
            data_filled_df(pd.DataFrame): Imputed DataFrame
    """
    if config.app.imputation_strategy in ["mean", "median", "most_frequent"]:
        # check if a timeseries pixel has only nans. Cant be filled
        # with these interpolation methods
        for _, one_band_df in samples_df.groupby(level=-1):
            if one_band_df.isnull().all(axis=0).any():
                TIMESERIES_LOGGER.error(
                    "Your data includes pixel which only contain NaN values "
                    "throughout the timeseries. Please choose one of the "
                    "following imputation methods: "
                    "'constant' or 'interpolate'")
                sys.exit(1)
    if config.app.imputation_strategy == 'interpolate':
        initial_columns = samples_df.columns
        samples_df.columns = [str(x) for x in range(len(samples_df.columns))]

        data_filled_df = samples_df.unstack(level=-1).interpolate(
            method='time',
            axis=0
        ).stack(dropna=False)
        data_filled_df.columns = initial_columns

        if data_filled_df.isnull().values.any():
            # If there are still any nodata pixels left...
            TIMESERIES_LOGGER.warning("Found pixels without any data. "
                                      "Setting mean chunk/roi values "
                                      "to the pixels")

            mean = data_filled_df.stack().mean()
            data_filled_df.fillna(mean, inplace=True)

    else:
        data_array = samples_df.values
        data_filled = [impute_values(data_array, config)]
        data_filled_df = pd.DataFrame(
            np.concatenate(data_filled),
            index=samples_df.index,
            columns=samples_df.columns)
    return data_filled_df


def get_timeseries_samples(rasters, rois, out_dir, config):
    """Samples for timeeseries from a list of rasters and a polygon file

        Args:
            rasters(list): list of raster files
            rois(str): Path to OGR polygon file
            out_dir (str): Path to save the samples to
            config (Configuration) contains config

        Returns:
            samples_df (pd.DataFrame): samples and classes
    """
    raster_dates = [get_raster_date(raster) for raster in rasters]
    raster_dict = dict(zip(raster_dates, rasters))
    if isinstance(rois, dict):
        samples_df = gather_samples_ts(raster_dict, config, roi=rois)
    else:
        samples_df = gather_samples_ts(raster_dict, config, rois_file=rois)
    if out_dir is not None:
        samples_df.to_csv(f'{out_dir}/samples_ts.csv')
    return samples_df



"The command line interface and entry point for classifier"
import logging
import os

import click

from classifier.accuracy import assess_accuracy
from classifier.classify import classify_, classify_timeseries
from classifier.settings import WORKSPACE
from classifier.utils.general import cli_init, save_config_as_json

CLI_LOGGER = logging.getLogger(__name__)


@click.group()
def cli():
    """---Classifier---
     Classification of rasters using ground truth"""


@cli.command()
@click.argument(
    'rasters',
    nargs=-1,
    type=click.Path(),
)
@click.option('--rois',
              default=None,
              type=click.Path(),
              help="""OGR supported polygon file""")
@click.option('--name',
              type=click.STRING,
              help="""Name of the output folder""")
@click.option('--overwrite',
              default=False,
              type=click.BOOL,
              help="""Overwrite Existing output folder"""
              )
@click.option('--config_location',
              default=os.path.join(WORKSPACE, 'config.json'),
              type=click.Path(),
              help="""Use a custom location for configuration file""")
def classification(name, rasters, rois, overwrite, config_location):
    """Do a traditional (un)supervised classification

    Raster input paths can be added like arguments without a prefix.
    Multiple raster paths can be added

    For additional options, create a config file in your workspace. Also see
    https://satelligence.gitlab.io/classifier/usage/configfile.html.

    """
    out_dir, rasters, rois, config = \
        cli_init(
            name,
            rasters,
            overwrite,
            config_location,
            rois
        )
    config.name = name
    if config.app.rasters_are_timeseries:
        classify_timeseries(rasters, config, out_dir, rois)
    else:
        classify_(rasters, config, out_dir, rois)


@ cli.command()
@ click.option('--name',
               type=click.STRING,
               help="""Name of the output folder""")
@ click.option('--raster',
               default=None,
               type=click.Path(),
               help="""Raster file with classification result""")
@ click.option('--rois',
               default=None,
               type=click.Path(),
               help="""OGR supported polygon file""")
@ click.option('--subset',
               default=None,
               type=click.Path(),
               help="""OGR supported polygon file""")
@ click.option('--overwrite',
               default=False,
               type=click.BOOL,
               help="""Overwrite Existing output folder""")
@ click.option('--config_location',
               default=os.path.join(WORKSPACE, 'config.json'),
               type=click.Path(),
               help="""Use a custom location for configuration file""")
def accuracy_assessment(name, raster, rois, subset, overwrite,
                        config_location):
    """Accuracy assessment.
    """
    out_dir, rasters, _, _ = cli_init(name, [raster], overwrite,
                                      config_location)

    cm_fn = os.path.join(out_dir, 'confusion_matrix_total')
    assess_accuracy(rasters[0], rois, cm_fn, subset)


@ cli.command()
def make_config():
    """Make a config file with default values in your workspace directory
    and exit"""
    save_config_as_json()


if __name__ == '__main__':
    cli()



"Supervised timeseries integration test for classifier"
from pathlib import Path

import numpy as np
import pytest
import rasterio
from classifier.cli import classification
from classifier.settings import WORKSPACE
from classifier.utils.general import save_config_as_json


@pytest.mark.parametrize('test_arg',
                         [[True, True]])
def test_supervised_timeseries_randomforest(runner):
    """
    Test for running a supervised randomforest classification
    on a timeseries
    """
    save_config_as_json(threads=1, algorithm="randomforest",
                        remove_outliers=False,
                        rasters_are_timeseries=True)

    rois_path = Path(WORKSPACE) / 'integration_test_rois.gpkg'
    runner.invoke(classification, [
        "--name", 'test_result', '--overwrite', True,
        '--rois', rois_path, "integration_test_rasters"])
    directory = Path(WORKSPACE) / 'test_result' / 'classification.tif'
    with rasterio.open(directory, 'r') as dst:
        result = dst.read(1)
        assert np.all(result[:, 0:4] == result[0, 0])
        assert np.all(result[:, 5:10] == result[0, 5])


@pytest.mark.parametrize('test_arg',
                         [[True, True]])
def test_supervised_timeseries_xgboost(runner):
    """
    Test for running a supervised xgboost classification
    on a timeseries
    """
    save_config_as_json(threads=1, algorithm="xgboost",
                        remove_outliers=False, rasters_are_timeseries=True)

    rois_path = Path(WORKSPACE) / 'integration_test_rois.gpkg'
    runner.invoke(classification, [
        "--name", 'test_result', '--overwrite', True,
        '--rois', rois_path, "integration_test_rasters"])
    directory = Path(WORKSPACE) / 'test_result' / 'classification.tif'
    with rasterio.open(directory, 'r') as dst:
        result = dst.read(1)
        assert np.all(result[:, 0:4] == result[0, 0])
        assert np.all(result[:, 5:10] == result[0, 5])



""" Unittests for config file"""
import copy
import json
import os
import unittest

from classifier.settings import PARAMETERS, WORKSPACE
from classifier.utils.config import (AccuracyConfiguration, AppConfiguration,
                                     Configuration,
                                     OptimizeSupervisedConfiguration,
                                     SupervisedConfiguration,
                                     UnsupervisedConfiguration, json_to_config)


class ConfigTestCase(unittest.TestCase):
    """ Testclass for config.py"""

    @classmethod
    def setUp(cls):
        algorithm = "randomforest"
        window = 1024
        model = None
        model_save = False  # Save a model file which can be re-ed
        samples = None
        log_level = "INFO"  # Logging level
        threads = -1  # of threads
        imputation = True  # e simple imputation for missing values
        imputation_strategy = "mean"  # Strategy for imputation. mean,
        rasters_are_timeseries = False  # Rasters are timeseries
        imputation_constant = -9999  # constant for imputation
        # if constant was chosen as imputation method
        probability = True
        all_probabilities = False
        optimize = False  # Optimize the model parameters
        optimize_number = 10  # Number of iterations for optimization
        max_features = ["auto", "sqrt", "log2"]
        max_leaf_nodes = [3, 5, 7]
        max_depth = [None, 1, 3, 10, 20000]
        n_estimators = [10, 50, 100]
        boxplots = False  # Plot Boxplots for samples
        remove_outliers = True  # Remove outliers from the training data
        nclasses = 2  # Number of classes for unsupervised
        trainfraction = 1.0  # Fraction of raster ed for training
        perform_assesment = True  # Perform accuracy assessment
        testfraction = 0.25  # Fraction of data to e for training

        cls.ref_config = Configuration(
            AppConfiguration(algorithm=algorithm,
                             window=window,
                             model=model,
                             model_save=model_save,
                             samples=samples,
                             log_level=log_level,
                             threads=threads,
                             imputation=imputation,
                             imputation_strategy=imputation_strategy,
                             rasters_are_timeseries=rasters_are_timeseries,
                             imputation_constant=imputation_constant),
            SupervisedConfiguration(
                probability=probability,
                all_probabilities=all_probabilities,
                boxplots=boxplots,
                remove_outliers=remove_outliers,
                optimization=OptimizeSupervisedConfiguration(
                    optimize=optimize,
                    optimize_number=optimize_number,
                    max_features=max_features,
                    max_leaf_nodes=max_leaf_nodes,
                    max_depth=max_depth,
                    n_estimators=n_estimators)),
            UnsupervisedConfiguration(nclasses, trainfraction),
            AccuracyConfiguration(perform_assesment, testfraction))

    def test_read_config_file_outputasexpected(self):
        """Config file is correctly read in"""
        params = copy.deepcopy(PARAMETERS)
        config = json_to_config(json.dumps(params))
        self.assertEqual(self.ref_config, config)

    def test_read_config_file__raises(self):
        """Raises error due to invalid config file"""
        params = copy.deepcopy(PARAMETERS)
        params['app']['model'] = "C:///:/Documen"

        with self.assertRaises(SystemExit):
            json_to_config(json.dumps(params))

    def test_read_config_file__model_not_zip_raises(self):
        """Raises error due to invalid config file
        model not .zip"""
        params = copy.deepcopy(PARAMETERS)
        params['app']['model'] = os.path.join(WORKSPACE, "model")

        with self.assertRaises(SystemExit):
            json_to_config(json.dumps(params))

    def test_read_config_file__samples_not_csv_or_pkl_raises(self):
        """Raises error due to invalid config file
        samples is not .pkl or .csv"""
        params = copy.deepcopy(PARAMETERS)
        params['app']['samples'] = os.path.join(WORKSPACE, "samples.jpg")

        with self.assertRaises(SystemExit):
            json_to_config(json.dumps(params))
